/* upscommon.c - functions used in more than one model support module

   Copyright (C) 1999  Russell Kroll <rkroll@exploits.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              
*/

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/termios.h>

#include "smartups.h"
#include "shared.h"
#include "config.h"
#include "version.h"
#include "upscommon.h"

	int	upsfd;
	char	statefn[256];
	itype	info[INFOMAX];
	
	int	flag_timeoutfailure = 0;

/* signal handler for SIGALRM when opening a serial port */
void openfail()
{
	printf ("Fatal error: serial port open timed out\n");
	exit (1);
}

/* open a serial port and setup the flags properly at a given speed */
void open_serial(char *port, speed_t speed)
{
	struct	termios	oldtio, newtio;
	int	res;

	signal (SIGALRM, openfail);
	alarm (3);

	upsfd = open (port, O_RDWR | O_NOCTTY | O_NONBLOCK);
	alarm (0);

	if (upsfd < 1) {
		printf ("Unable to open %s: %s\n", port, strerror(errno));
		exit (1);
	}

	res = 0;	/* quiet gcc */
#ifdef HAVE_FLOCK
	if (flock(upsfd, LOCK_EX | LOCK_NB) != 0) {
		printf ("%s is locked by another process\n", port);
		exit (1);
	}
#endif
#ifdef HAVE_UULOCK
	res = uu_lock(port);
	if (res != 0) {
		printf ("Can't lock %s: %s\n", port, uu_lockerr(res));
		exit (1);
	}
#endif

	tcgetattr (upsfd, &oldtio);
	newtio.c_lflag = 0 | ECHOE | ECHOKE | ECHOCTL | PENDIN;
	newtio.c_iflag = 0 | IXANY | IMAXBEL | IXOFF;
	newtio.c_oflag = 0 | ONLCR;
	newtio.c_cflag = 0 | CREAD | CS8 | HUPCL | CLOCAL;
	cfsetospeed (&newtio, speed);
	cfsetispeed (&newtio, speed);

	newtio.c_cc[VMIN] = 1;
	newtio.c_cc[VTIME] = 0;
	tcflush (upsfd, TCIFLUSH);
	tcsetattr(upsfd, TCSANOW, &newtio);
	close (upsfd);

	/* reopen for normal use */

	signal (SIGALRM, openfail);
	alarm (3);

	upsfd = open (port, O_RDWR | O_NOCTTY);
	alarm (0);

	res = 0;	/* quiet gcc */
#ifdef HAVE_FLOCK
	if (flock(upsfd, LOCK_EX | LOCK_NB) != 0) {
		printf ("%s is locked by another process\n", port);
		exit (1);
	}
#endif
#ifdef HAVE_UULOCK
	res = uu_lock(port);
	if (res != 0) {
		printf ("Can't lock %s: %s\n", port, uu_lockerr(res));
		exit (1);
	}
#endif

	if (upsfd < 1) {
		printf ("Unable to open %s: %s\n", port, strerror(errno));
		exit (1);
	}

	signal (SIGALRM, SIG_IGN);

	tcgetattr (upsfd, &oldtio);
	newtio.c_cflag = speed | CS8 | CLOCAL | CREAD;
	newtio.c_iflag = IGNPAR;
	newtio.c_oflag = 0;
	newtio.c_lflag = 0;
	newtio.c_cc[VMIN] = 1;
	newtio.c_cc[VTIME] = 0;

	tcflush (upsfd, TCIFLUSH);
	tcsetattr (upsfd, TCSANOW, &newtio);
}

/* put a notice in the syslog */
void notice (char *msg)
{
	char	logbuf[256];

	snprintf (logbuf, sizeof(logbuf), "Notice: %s", msg);
	syslog (LOG_NOTICE, msg);
}

/* put a fatal error in the syslog and exit */
void fatal (char *msg)
{
	char	logbuf[256];

	snprintf (logbuf, sizeof(logbuf), "Fatal error: %s", msg);
	syslog (LOG_ERR, msg);
	exit (1);
}

/* function for erasing "timeout"-conditions */
void nolongertimeout()
{
	static int flag_firstcall = 0;

	if (flag_firstcall == 0) {
	    notice ("Serial port read ok");
	    flag_firstcall = 1;
	}
	
	if (flag_timeoutfailure == 1) {
	    notice ("Serial port read ok again");
	}
	
	flag_timeoutfailure = 0;
	return;
}


/* signal handler for SIGALRM when trying to read */
void timeout()
{
	struct sigaction sa;
	sigset_t sigmask;

	if (flag_timeoutfailure == 0) {
	    notice ("Serial port read timed out");
	}
	
	flag_timeoutfailure = 1;
	
	sa.sa_handler = SIG_DFL;
	sigemptyset (&sigmask);
	sa.sa_mask = sigmask;
	sa.sa_flags = 0;
	sigaction (SIGALRM, &sa, NULL);
	return;
}

/* wait for an answer in the form <data><CR><LF> and store in buf */
int recv (char *buf, int buflen)
{
	char recvbuf[512], in;
	int ret;
	struct sigaction sa;
	sigset_t sigmask;

	strcpy (recvbuf, "");

	sa.sa_handler = timeout;
	sigemptyset (&sigmask);
	sa.sa_mask = sigmask;
	sa.sa_flags = 0;
	sigaction (SIGALRM, &sa, NULL);

	alarm (3);

	for (;;) {
		ret = read (upsfd, &in, 1);
		if (ret > 0) {
			switch (in) {
				case 10:
					alarm (0);
					signal (SIGALRM, SIG_IGN);
					strncpy (buf, recvbuf, buflen);
					buf[buflen - 1] = 0;
					return (buflen);
					break;
				case 13:	/* ignore CR */
				case '!':	/* TODO: APC special chars */
				case '$':	/* ditto */
					break;
				default:
					sprintf (recvbuf, "%s%c", recvbuf, in);
					break;
			}
			nolongertimeout();
		}
		else {
			return (-1);
		}

		/* keep from overrunning the buffer - lame hack */
		if (strlen (recvbuf) > 256) {
			notice ("UPS is spewing wildly");
			strcpy (recvbuf, "");
		}
	}

	return (-1);
}

/* send a single byte to the UPS */
int send (char data)
{
	tcflush (upsfd, TCIFLUSH);
	return (write (upsfd, &data, 1));
}

/* get data from the UPS and install it in the data array */
void installinfo (int infotype, char reqchar)
{
	int	i, pos = 0;

	for (i = 0; i < INFOMAX; i++)
		if (info[i].type == infotype) {
			pos = i;
			break;
		}

	if (pos == 0) {	/* not found, probably debugging? */
		printf ("installinfo: can't find type %i\n", infotype);
		return;
	}

	send (reqchar);
	recv (info[pos].value, sizeof(info[pos].value));
}

/* find data of a given type in the info array */
char *getdata (int infotype)
{
	int	i;

	for (i = 0; i < INFOMAX; i++)
		if (info[i].type == infotype)
			return (info[i].value);
	
	return (NULL);
}

/* write data to state file - effectively a no-op for shared memory mode */
void writeinfo (int initial)
{
	int	sffd, ret;

	/* TODO: detect shm mode and return immediately */

	/* if data is stale, don't write it so the state file ages */
	if (flag_timeoutfailure == 1)
		return;

	sffd = open (statefn, O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR);
	if (sffd < 0)
		if (initial) {
			if (errno == EACCES)
				printf ("This file has to be accessible by userid %i\n", RUN_AS_UID);
			printf ("Can't open %s: %s\n", statefn, strerror(errno));
			exit (1);
		}

	ret = write (sffd, &info, sizeof(itype) * INFOMAX);
	if (ret < 0)
		if (initial) {
			printf ("Can't open %s: %s\n", statefn, strerror(errno));
			if (errno == EPERM)
				printf ("This file has to be writable by userid %i\n", RUN_AS_UID);
			exit (1);
		}

	close (sffd);
}

void background()
{
	int	pid;
	
	if ((pid = fork()) < 0) {
		perror ("Unable to enter background");
		exit (1);
	}

	close (0);
	close (1);
	close (2);

	if (pid != 0) 
		exit (0);		/* parent */

	/* child */

	syslog (LOG_INFO, "Startup successful");
}

void droproot ()
{
	setgid(RUN_AS_GID);

	if (getuid() == 0)
		if (setuid(RUN_AS_UID) == -1) {
			perror ("setuid");
			exit (1);
		}
}
